#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "windowproperties.h"

#include <QDebug>
#include <QImage>
#include <QPixmap>
#include <QTreeWidgetItem>
#include <QMessageBox>
#include <QX11Info>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    totalWindows(0)
{
    ui->setupUi(this);

    // for XmbTextPropertyToTextList
    setlocale(LC_ALL, ""); // see man locale

    refreshList();
}

MainWindow::~MainWindow()
{
    delete ui;
}

const auto XWindowIDRole = Qt::UserRole + 1;
#define SET_WINDOW(widget, wnd) widget->setData(0, XWindowIDRole, static_cast<qlonglong>(wnd))
#define GET_WINDOW(widget) static_cast<Window>( widget->data(0, XWindowIDRole).value<qlonglong>() );

static bool windowClass(Window w, QString& setName, QString& setClass) {
    XClassHint *c = XAllocClassHint(); // see man
    if (c == nullptr) {
        return "";
    }

    Status s = XGetClassHint(QX11Info::display(), w, c); // see man
    if (s >= 0) {
        setName = c->res_name;
        setClass = c->res_class;
    }

    XFree(c);
    return (s >= 0);
}

static bool windowTitle(Window w, QString& setTitle) {

    XTextProperty text_prop;
    Status s = XGetWMName(QX11Info::display(), w, &text_prop);
    if (!s || !text_prop.value || !text_prop.nitems) {
        return false;
    }

    char **list = nullptr;
    int num = 0;
    s = XmbTextPropertyToTextList(QX11Info::display(), &text_prop, &list, &num);
    if (s < Success || !num || !*list) {
        return false;
    }

    setTitle = QString(*list);

    XFree(text_prop.value);
    XFreeStringList(list);
    return true;
}

#define COLUMN_ID    0
#define COLUMN_NAME  1
#define COLUMN_CLASS 2
#define COLUMN_TITLE 3

static void applyItemProperties(QTreeWidgetItem *itm, Window wnd) {

    itm->setText(COLUMN_ID, QString("%1").arg(wnd, 8, 16));

    QString title, name, classname;
    if (windowTitle(wnd, title)) {
        itm->setText(COLUMN_TITLE, title);
    }

    if (windowClass(wnd, name, classname)) {
        itm->setText(COLUMN_NAME, name);
        itm->setText(COLUMN_CLASS, classname);
    }

    //
    //rootItem->setText(0, "(root)");
    //SET_WINDOW(rootItem, rootWnd);

    SET_WINDOW(itm, wnd);
}

static unsigned int refreshChildren(QTreeWidgetItem* parentItem, Window parentWindow) {
    // todo
    Window *children = nullptr;
    Window root;
    Window parent;
    unsigned int nChildren = 0;

    Status s = XQueryTree(QX11Info::display(), parentWindow, &root, &parent, &children, &nChildren);
    if (s < 0) {
        return 0;
    }

    unsigned int totalChildrenFound = nChildren;
    for (unsigned int i = 0; i < nChildren; ++i) {

        auto *itm = new QTreeWidgetItem();
        applyItemProperties(itm, children[i]);
        parentItem->addChild(itm);

        totalChildrenFound += refreshChildren(itm, children[i]);
    }

    XFree(children);
    return totalChildrenFound;
}

void MainWindow::refreshList()
{
    totalWindows = 1;

    Window rootWnd = static_cast<Window>(QX11Info::appRootWindow()); // or: XDefaultRootWindow(QX11Info::display());

    ui->windowTree->setEnabled(false);
    ui->windowTree->clear();

    auto *rootItem = new QTreeWidgetItem();
    applyItemProperties(rootItem, rootWnd);
    ui->windowTree->addTopLevelItem(rootItem);

    totalWindows += refreshChildren(rootItem, rootWnd);

    ui->windowTree->expandAll();
    ui->windowTree->setEnabled(true);
    on_windowTree_itemSelectionChanged();

    setWindowTitle(tr("%1 (%2)").arg(QApplication::applicationDisplayName()).arg(totalWindows));
}

void MainWindow::on_actionRefresh_triggered()
{
    refreshList();
}

void MainWindow::on_actionAbout_X11Glove_triggered()
{
    QMessageBox::about(
        this, tr("About %1").arg(QApplication::applicationDisplayName()),

        "<b>About X11Glove</b><br>"
        "<br>"
        "Copyright (2017) The X11Glove Author(s)<br>"
        "<br>"
        "- <a href=\"https://code.ivysaur.me/x11glove\">Homepage</a><br>"
        "- <a href=\"https://tronche.com/gui/x/xlib/\">The Xlib Programming Manual</a><br>"
    );
}

void MainWindow::on_actionAbout_Qt_triggered()
{
    QApplication::aboutQt();
}

void MainWindow::on_actionExit_triggered()
{
    close();
}

void MainWindow::on_windowTree_itemSelectionChanged()
{
    if (ui->windowTree->selectedItems().size() == 0) {
        ui->tabWidget->setVisible(false);
        ui->actionClear_selection->setEnabled(false);
        return;
    }

    // Item selected
    ui->tabWidget->setVisible(true);
    ui->actionClear_selection->setEnabled(true);

    // Reload the property list for the selected item
    renderDetailsForSelectedWindow();
}

void MainWindow::renderDetailsForSelectedWindow()
{
    Window w = GET_WINDOW( ui->windowTree->selectedItems()[0] );
    renderDetailsFor(w);
}

QString MainWindow::formatGravity(int gravity) {
    switch(gravity) {
        case ForgetGravity:    return "ForgetGravity / UnmapGravity";
        //case UnmapGravity:     return "UnmapGravity"; // duplicate #define value
        case StaticGravity:    return "StaticGravity";
        case CenterGravity:    return "CenterGravity";
        case NorthWestGravity: return "NorthWestGravity";
        case NorthGravity:     return "NorthGravity";
        case NorthEastGravity: return "NorthEastGravity";
        case EastGravity:      return "EastGravity";
        case SouthEastGravity: return "SouthEastGravity";
        case SouthGravity:     return "SouthGravity";
        case SouthWestGravity: return "SouthWestGravity";
        case WestGravity:      return "WestGravity";

    default:
        return tr("Unknown (%1)").arg(gravity);
    }
}

void MainWindow::renderDetailsFor(Window w)
{
    // ATTRIBUTES

    XWindowAttributes wa;
    XGetWindowAttributes(QX11Info::display(), w, &wa);
    ui->attrPosition   ->setText( tr("%1, %2").arg(wa.x).arg(wa.y) );
    ui->attrDimensions ->setText( tr("%1 x %2").arg(wa.width).arg(wa.height) );
    ui->attrBorderWidth->setText( tr("%1 px").arg(wa.border_width) );
    ui->attrDepth      ->setText( tr("%1 bits per pixel").arg(wa.depth) );
    if (wa.c_class == InputOnly) {
        ui->attrClass->setText("InputOnly");
    } else if (wa.c_class == InputOutput) {
        ui->attrClass->setText("InputOutput");
    } else {
        ui->attrClass->setText(tr("(invalid)"));
    }
    ui->attrBitGravity->setText(formatGravity(wa.bit_gravity));
    ui->attrWindowGravity->setText(formatGravity(wa.win_gravity));
    if (wa.backing_store == WhenMapped) {
        ui->attrBackingStore->setText("WhenMapped");
    } else if (wa.backing_store == Always) {
        ui->attrBackingStore->setText("Always");
    } else if (wa.backing_store == NotUseful) {
        ui->attrBackingStore->setText("NotUseful");
    } else {
        ui->attrBackingStore->setText(tr("(invalid)"));
    }
    // TODO there are several more properties here that can be displayed
    // @ref https://tronche.com/gui/x/xlib/window-information/XGetWindowAttributes.html

    // PROPERTIES

    ui->propertyList->clear();

    WindowProperties wp(w);
    auto props = wp.CurrentProperties();
    for (auto i = props.begin(); i != props.end(); ++i) {
        QTreeWidgetItem *itm = new QTreeWidgetItem();

        itm->setText(0, i.key());
        itm->setText(1, QString::fromUtf8(i.value().toByteArray()));

        ui->propertyList->addTopLevelItem(itm);
    }

    // IMAGE PREVIEW

    XImage *image = XGetImage(QX11Info::display(), w, 0, 0, wa.width, wa.height, AllPlanes, ZPixmap);
    if (image == nullptr) {
        ui->windowPreview->clear();
        ui->windowPreview->setText(tr("No preview available."));

    } else {

        QImage im;
        if (wa.depth == 32) {
            im = QImage(wa.width, wa.height, QImage::Format_ARGB32);
            for (int i = 0; i < wa.height; ++i) {
                for (int j = 0; j < wa.width; ++j) {
                    unsigned long pixel = XGetPixel(image, j, i);
                    im.setPixel(j, i, pixel);
                }
            }
        } else if (wa.depth == 24) {
            im = QImage(wa.width, wa.height, QImage::Format_RGBX8888);
            for (int i = 0; i < wa.height; ++i) {
                for (int j = 0; j < wa.width; ++j) {
                    unsigned long pixel = XGetPixel(image, j, i);
                    im.setPixel(j, i, pixel);
                }
            }
        }

        XDestroyImage(image);

        ui->windowPreview->clear();
        ui->windowPreview->setPixmap(QPixmap::fromImage(im));
    }
}

void MainWindow::on_actionClear_selection_triggered()
{
    ui->windowTree->clearSelection();
}

void MainWindow::on_lineEdit_textChanged(const QString &arg1)
{
    Q_UNUSED(arg1);
    ui->windowTree->clearSelection();
}
